Imports System.Device.Location
Imports Microsoft.Devices.Sensors
Imports Petzold.Astronomy
Imports Petzold.Phone.Xna

Namespace ViewThreeCoordinates
	Public Class Game1
		Inherits Microsoft.Xna.Framework.Game

		Private Const WIDTH As Integer = 480
		Private Const HEIGHT As Integer = 800
		Private Const PIXELS_PER_DEGREE As Integer = 15
        Private Shared ReadOnly DIVISION_ANGLE As Angle = Angle.FromDegrees(15)
        Private Shared ReadOnly numAzimuthAngles As Integer = CInt(Angle.Circle / DIVISION_ANGLE)
        Private Shared ReadOnly numPolarAngles As Integer = CInt(Angle.Straight / DIVISION_ANGLE) + 1

		Private graphics As GraphicsDeviceManager
        Private spriteBatch As SpriteBatch

		' Graphics relatied
		Private labelFont As SpriteFont
		Private displayMatrix As Matrix = Matrix.Identity
		Private lineRenderer As LineRenderer
		Private coordinateProjection As New HorizontalCoordinateProjection

		' Sensors
		Private motion As Motion
		Private geoWatcher As GeoCoordinateWatcher

		' Displayed data
		Private horizontalScreenPoints(numAzimuthAngles - 1, numPolarAngles - 1) As Vector2
		Private equatorialScreenPoints(numAzimuthAngles - 1, numPolarAngles - 1) As Vector2
		Private eclipticScreenPoints(numAzimuthAngles - 1, numPolarAngles - 1) As Vector2

		Private Class Label
			Public Sub New(ByVal text As String)
				Me.Text = text
			End Sub

			Public Text As String
			Public Size As Vector2
			Public Position As Vector2
		End Class

		Private compassLabels() As Label = { New Label("North"), New Label("NE"), New Label("East"), New Label("SE"), New Label("South"), New Label("SW"), New Label("West"), New Label("NW") }

		Private rightAscensionLabels(23) As Label

        Private zodiacLabels() As Label = {New Label("0°"), New Label("Aries"), New Label("30°"), New Label("Taurus"), New Label("60°"), New Label("Gemini"), New Label("90°"), New Label("Cancer"), New Label("120°"), New Label("Leo"), New Label("150°"), New Label("Virgo"), New Label("180°"), New Label("Libra"), New Label("210°"), New Label("Scorpio"), New Label("240°"), New Label("Sagittarius"), New Label("270°"), New Label("Capricorn"), New Label("300°"), New Label("Aquarius"), New Label("330°"), New Label("Pisces")}

		Private altitudeLabels(numPolarAngles - 2 - 1) As Label

		Private visibilityMode As Integer = 6

		Public Sub New()
			graphics = New GraphicsDeviceManager(Me)
			Content.RootDirectory = "Content"

			' Frame rate is 30 fps by default for Windows Phone.
			TargetElapsedTime = TimeSpan.FromTicks(333333)

			' Extend battery life under lock.
			InactiveSleepTime = TimeSpan.FromSeconds(1)

			' Switch to portrait mode
			graphics.IsFullScreen = True
			graphics.PreferredBackBufferWidth = WIDTH
			graphics.PreferredBackBufferHeight = HEIGHT

			' Create Motion object
			If Motion.IsSupported Then
				motion = New Motion With {.TimeBetweenUpdates = Me.TargetElapsedTime}
			End If

			' Create GPS watcher
			geoWatcher = New GeoCoordinateWatcher
		End Sub

		Protected Overrides Sub OnActivated(ByVal sender As Object, ByVal args As EventArgs)
			If motion IsNot Nothing Then
				Try
					motion.Start()
				Catch
				End Try
			End If

			geoWatcher.Start()
			MyBase.OnActivated(sender, args)
		End Sub

		Protected Overrides Sub OnDeactivated(ByVal sender As Object, ByVal args As EventArgs)
			If motion IsNot Nothing Then
				motion.Stop()
			End If

			geoWatcher.Stop()
			MyBase.OnDeactivated(sender, args)
		End Sub

		Protected Overrides Sub Initialize()
			' Initialize some labels
			For index  = 0 To altitudeLabels.Length - 1
				altitudeLabels(index) = New Label(((index + 1) * DIVISION_ANGLE + -Angle.Right).ToString)
			Next index

			For index  = 0 To rightAscensionLabels.Length - 1
                Dim hour As Double = CLng(24.0 * index) \ rightAscensionLabels.Length
				rightAscensionLabels(index) = New Label(hour.ToString & " h")
			Next index

			' Enable taps 
			TouchPanel.EnabledGestures = GestureType.Tap
			MyBase.Initialize()
		End Sub

		Protected Overrides Sub LoadContent()
			' Create a new SpriteBatch, which can be used to draw textures.
			spriteBatch = New SpriteBatch(GraphicsDevice)

			' Load the font and obtain text sizes
			labelFont = Me.Content.Load(Of SpriteFont)("LabelFont")

            For Each label In compassLabels
                label.Size = labelFont.MeasureString(label.Text)
            Next label

            For Each label In altitudeLabels
                label.Size = labelFont.MeasureString(label.Text)
            Next label

            For Each label In rightAscensionLabels
                label.Size = labelFont.MeasureString(label.Text)
            Next label

            For Each label In zodiacLabels
                label.Size = labelFont.MeasureString(label.Text)
            Next label

			' Create the line-rendering class
			lineRenderer = New LineRenderer(Me.GraphicsDevice)
		End Sub

		Protected Overrides Sub Update(ByVal gameTime As GameTime)
			' Allows the game to exit
			If GamePad.GetState(PlayerIndex.One).Buttons.Back = ButtonState.Pressed Then
				Me.Exit()
			End If

			' If screen tapped, change visibility mode
			Do While TouchPanel.IsGestureAvailable
				If TouchPanel.ReadGesture.GestureType = GestureType.Tap Then
					visibilityMode += 1
				End If
			Loop
					visibilityMode = visibilityMode Mod 7

			If motion IsNot Nothing AndAlso motion.IsDataValid Then
				' Create a HorizontalCoordinate from the 3D rotation matrix
				Dim matrix As Matrix = motion.CurrentValue.Attitude.RotationMatrix
				Dim viewCenter As HorizontalCoordinate = HorizontalCoordinate.From(matrix)

				' Use tilt to globally rotate the graphics
				displayMatrix = Matrix.CreateTranslation(-WIDTH \ 2, -HEIGHT \ 2, 0) * Matrix.CreateRotationZ(CSng(-viewCenter.Tilt.Radians)) * Matrix.CreateTranslation(WIDTH \ 2, HEIGHT \ 2, 0)

				' Set the view center to the HorizontalCoordinateProjection
				coordinateProjection.SetViewCenter(viewCenter)

				' Get the current time
				Dim time As Time = Time.GetNow

				' Get the current location
				Dim location As GeoCoordinate = geoWatcher.Position.Location
				Dim geographicCoordinate As New GeographicCoordinate(Angle.FromDegrees(-location.Longitude), Angle.FromDegrees(location.Latitude))

				For azimuthIndex  = 0 To horizontalScreenPoints.GetLength(0) - 1
					For altitudeIndex  = 0 To horizontalScreenPoints.GetLength(1) - 1
						' Calculate positions of horizontal coordinates
						Dim azimuth As Angle = DIVISION_ANGLE * azimuthIndex
						Dim altitude As Angle = -Angle.Right + DIVISION_ANGLE * altitudeIndex
						Dim horizontalCoordinate As New HorizontalCoordinate(azimuth, altitude)
						Dim point As Vector2 = CalculateScreenPoint(horizontalCoordinate)
						horizontalScreenPoints(azimuthIndex, altitudeIndex) = point

						' Calculate positions of equatorial coordinates
						Dim rightAscension As Angle = DIVISION_ANGLE * azimuthIndex
						Dim declination As Angle = -Angle.Right + DIVISION_ANGLE * altitudeIndex
						Dim equatorialCoordinate As New EquatorialCoordinate(rightAscension, declination)
						horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time)
						point = CalculateScreenPoint(horizontalCoordinate)
						equatorialScreenPoints(azimuthIndex, altitudeIndex) = point

						' Calculate positions of ecliptic coordinates
						Dim longitude As Angle = DIVISION_ANGLE * azimuthIndex
						Dim latitude As Angle = -Angle.Right + DIVISION_ANGLE * altitudeIndex
						Dim eclipticCoordinate As New EclipticCoordinate(longitude, latitude, 0)
						equatorialCoordinate = EquatorialCoordinate.From(eclipticCoordinate, time)
						horizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time)
						point = CalculateScreenPoint(horizontalCoordinate)
						eclipticScreenPoints(azimuthIndex, altitudeIndex) = point
					Next altitudeIndex
				Next azimuthIndex

				' Calculate positions of compass points
				For index  = 0 To compassLabels.Length - 1
					Dim label As Label = compassLabels(index)
					Dim azimuth As Angle = index * Angle.Circle / compassLabels.Length
					Dim altitude As Angle = Angle.FromDegrees(Math.Min(80, Math.Max(-80, viewCenter.Altitude.Degrees)))
					Dim horizontalCoordinate As New HorizontalCoordinate(azimuth, altitude)
					Dim point As Vector2 = CalculateScreenPoint(horizontalCoordinate)
					label.Position = New Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2)
				Next index

				' Calculate positions of altitude labels
				For index  = 0 To altitudeLabels.Length - 1
					Dim label As Label = altitudeLabels(index)
					Dim altitude As Angle = (index + 1) * DIVISION_ANGLE - Angle.Right
					Dim horizontalCoordinate As New HorizontalCoordinate(viewCenter.Azimuth, altitude)
					Dim point As Vector2 = CalculateScreenPoint(horizontalCoordinate)
					label.Position = New Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2)
				Next index

				' Calculate positions of right ascension labels
				For index  = 0 To rightAscensionLabels.Length - 1
					Dim label As Label = rightAscensionLabels(index)
					Dim rightAscension As Angle = index * Angle.Circle / rightAscensionLabels.Length
					Dim equatorialCoordinate As New EquatorialCoordinate(rightAscension, Angle.Zero)
					Dim horizontalCoordinate As HorizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time)
					Dim point As Vector2 = CalculateScreenPoint(horizontalCoordinate)
					label.Position = New Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2)
				Next index

				' Calculate positions of zodiac labels
				For index  = 0 To zodiacLabels.Length - 1
					Dim label As Label = zodiacLabels(index)
					Dim longitude As Angle = index * Angle.Circle / zodiacLabels.Length
					Dim eclipticCoordinate As New EclipticCoordinate(longitude, Angle.Zero, 0)
					Dim equatorialCoordinate As EquatorialCoordinate = EquatorialCoordinate.From(eclipticCoordinate, time)
					Dim horizontalCoordinate As HorizontalCoordinate = HorizontalCoordinate.From(equatorialCoordinate, geographicCoordinate, time)
					Dim point As Vector2 = CalculateScreenPoint(horizontalCoordinate)
					label.Position = New Vector2(point.X - label.Size.X / 2, point.Y - label.Size.Y / 2)
				Next index
			End If
			MyBase.Update(gameTime)
		End Sub

		Private Function CalculateScreenPoint(ByVal horizontalCoordinate As HorizontalCoordinate) As Vector2
			Dim horzAngle, vertAngle As Angle
			coordinateProjection.GetAngleOffsets(horizontalCoordinate, horzAngle, vertAngle)

			' Use NaN to indicate points clearly out of range of the screen
			Dim x As Single = Single.NaN
			Dim y As Single = Single.NaN

			If horzAngle > -Angle.Right AndAlso horzAngle < Angle.Right AndAlso vertAngle > -Angle.Right AndAlso vertAngle < Angle.Right Then
				x = CSng(WIDTH \ 2 + PIXELS_PER_DEGREE * horzAngle.Degrees)
				y = CSng(HEIGHT \ 2 + PIXELS_PER_DEGREE * vertAngle.Degrees)
			End If
			Return New Vector2(x, y)
		End Function

		Protected Overrides Sub Draw(ByVal gameTime As GameTime)
			GraphicsDevice.Clear(New Color(0, 0, 32))

			spriteBatch.Begin(SpriteSortMode.Immediate, Nothing, Nothing, Nothing, Nothing, Nothing, displayMatrix)

			If ((visibilityMode + 1) And 1) <> 0 Then
				DrawCoordinates(horizontalScreenPoints, compassLabels, altitudeLabels, Color.Green)
			End If

			If ((visibilityMode + 1) And 2) <> 0 Then
				DrawCoordinates(equatorialScreenPoints, rightAscensionLabels, Nothing, Color.Blue)
			End If

			If ((visibilityMode + 1) And 4) <> 0 Then
				DrawCoordinates(eclipticScreenPoints, zodiacLabels, Nothing, Color.Red)
			End If

			spriteBatch.End()
			MyBase.Draw(gameTime)
		End Sub

		Private Sub DrawCoordinates(ByVal screenPositions(,) As Vector2, ByVal azimuthLabels() As Label, ByVal polarAngleLabels() As Label, ByVal color As Color)
			Dim numAzimuths As Integer = screenPositions.GetLength(0)
			Dim numAltitudes As Integer = screenPositions.GetLength(1)

			' Draw circles of altitude
			For altitudeIndex  = 1 To numAltitudes - 2
				Dim thickness As Single = If(altitudeIndex = numAltitudes \ 2, 5, 2)

				For azimuthIndex  = 0 To numAzimuths - 1
					Dim point1 As Vector2 = screenPositions(azimuthIndex, altitudeIndex)
					Dim point2 As Vector2 = screenPositions((azimuthIndex + 1) Mod numAzimuths, altitudeIndex)

					If (Not Single.IsNaN(point1.X)) AndAlso (Not Single.IsNaN(point2.X)) Then
						lineRenderer.Draw(spriteBatch, point1, point2, thickness, color)
					End If
				Next azimuthIndex
			Next altitudeIndex

			' Draw azimuth lines
			For azimuthIndex  = 0 To numAzimuths - 1
				For altitudeIndex  = 0 To numAltitudes - 2
					Dim point1 As Vector2 = screenPositions(azimuthIndex, altitudeIndex)
					Dim point2 As Vector2 = screenPositions(azimuthIndex, altitudeIndex + 1)

					If (Not Single.IsNaN(point1.X)) AndAlso (Not Single.IsNaN(point2.X)) Then
						lineRenderer.Draw(spriteBatch, point1, point2, 2, color)
					End If
				Next altitudeIndex
			Next azimuthIndex

			' Draw azimuth labels
			If azimuthLabels IsNot Nothing Then
                For Each label In azimuthLabels
                    If Not Single.IsNaN(label.Position.X) Then
                        spriteBatch.DrawString(labelFont, label.Text, label.Position, color)
                    End If
                Next label
			End If

			' Draw polar angle labels
			If polarAngleLabels IsNot Nothing Then
                For Each label In polarAngleLabels
                    If Not Single.IsNaN(label.Position.X) Then
                        spriteBatch.DrawString(labelFont, label.Text, label.Position, color)
                    End If
                Next label
			End If
		End Sub
	End Class
End Namespace
